# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-05-13 11:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0002_messageset_content_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='messageset',
            name='short_name',
            field=models.CharField(max_length=100, unique=True, verbose_name='Short name'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-11-02 09:50
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0007_auto_20171102_0950'),
        ('subscriptions', '0006_auto_20170727_0820'),
    ]

    operations = [
        migrations.CreateModel(
            name='EstimatedSend',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('send_date', models.DateField()),
                ('estimate_subscriptions', models.IntegerField()),
                ('estimate_identities', models.IntegerField()),
                ('messageset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='estimates', to='contentstore.MessageSet')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='estimatedsend',
            unique_together=set([('send_date', 'messageset')]),
        ),
    ]



from django.core.management.base import BaseCommand
from django.core.validators import URLValidator
from django.conf import settings

from subscriptions.models import Subscription
from seed_services_client import SchedulerApiClient


def url_validator(url_str):
    URLValidator(url_str)
    return url_str


class Command(BaseCommand):
    help = 'Removes all duplicate subscriptions'

    DEFAULT_TIME_DELTA = 60
    DEFAULT_FIELDS = [
        'identity', 'version', 'messageset', 'lang', 'active',
        'completed', 'schedule']

    def add_arguments(self, parser):
        parser.add_argument(
            '--time-delta', action='store', dest='time_delta', type=int,
            nargs='?', default=self.DEFAULT_TIME_DELTA,
            help='The maximum time difference in seconds between two '
            'subscriptions for them to be considered duplicates.'
        )
        parser.add_argument(
            '--fields', action='store', dest='fields', type=str, nargs='+',
            default=self.DEFAULT_FIELDS,
            help='The fields to compare to determine if the subscriptions are'
            ' identical or not.'
        )
        parser.add_argument(
            '--fix', action='store_true', default=False,
            help=('Actually remove the duplicates instead of only listing'
                  'duplicates')
        )
        parser.add_argument(
            '--scheduler-url',
            type=url_validator,
            default=settings.SCHEDULER_URL,
            help='The scheduler url to use, defaults to settings.SCHEDULER_URL'
        )
        parser.add_argument(
            '--scheduler-token',
            type=str,
            default=settings.SCHEDULER_API_TOKEN,
            help=('The schedule API token, defaults to '
                  'settings.SCHEDULER_API_TOKEN')
        )

    def second_diff(self, d1, d2):
        '''
        Returns the difference, in seconds, between d1 and d2. The amount
        returned is always positive.
        '''
        return abs((d1 - d2).total_seconds())

    def is_within_limits(self, limit, date, dates):
        '''
        Returns True if the difference between date and any value in dates
        is less than or equal to limit.
        '''
        return any((self.second_diff(date, d) <= limit for d in dates))

    def handle(self, *args, **options):
        time_delta = options['time_delta']
        fields = options['fields']
        fix = options['fix']
        scheduler_token = options['scheduler_token']
        scheduler_url = options['scheduler_url']
        removed = 0

        scheduler_client = SchedulerApiClient(scheduler_token, scheduler_url)

        uniques = Subscription.objects.distinct(*fields)
        for unique in uniques:
            subscriptions = Subscription.objects.filter(**{
                field: getattr(unique, field) for field in fields}
            )
            subscriptions = subscriptions.filter(active=True, completed=False)
            subscriptions = subscriptions.order_by('created_at')
            if len(subscriptions) <= 1:
                continue
            dates = [subscriptions[0].created_at]
            for sub in subscriptions[1:]:
                if self.is_within_limits(time_delta, sub.created_at, dates):
                    if fix:
                        scheduler_id = sub.get_scheduler_id()
                        # NOTE: grabbing the signature of the sub before
                        #       deleting otherwise the warning below just
                        #       shows `None` as Django deletes the PK
                        #       after delete() is called
                        sub_str = str(sub)
                        sub.delete()
                        if scheduler_id:
                            scheduler_client.delete_schedule(scheduler_id)
                        else:
                            self.warning(
                                'Subscription %s has no scheduler_id.' % (
                                    sub_str,))
                    else:
                        self.warning(
                            'Not removing %s, use --fix to actually remove.' %
                            (sub,))

                    removed += 1
                dates.append(sub.created_at)

        self.success('Removed %d duplicate subscriptions.' % (removed,))

    def log(self, level, msg):
        self.stdout.write(level(msg))

    def warning(self, msg):
        self.log(self.style.WARNING, msg)

    def success(self, msg):
        self.log(self.style.SUCCESS, msg)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-09-12 09:23
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0004_auto_20160810_0916'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='message',
            unique_together=set([('messageset', 'sequence_number', 'lang')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-11-02 09:50
from __future__ import unicode_literals

import contentstore.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0006_messageset_channel'),
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='text_content',
            field=models.TextField(blank=True, null=True, validators=[contentstore.models.validate_special_characters]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-22 15:34
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0002_messageset_content_type'),
        ('subscriptions', '0002_auto_20160224_0822'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='subscription',
            name='messageset_id',
        ),
        migrations.AddField(
            model_name='subscription',
            name='messageset',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='contentstore.MessageSet'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='subscription',
            name='created_by',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions_created', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='schedule',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='contentstore.Schedule'),
        ),
        migrations.AlterField(
            model_name='subscription',
            name='updated_by',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions_updated', to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.contrib import admin
from django.contrib import messages
from django.shortcuts import render
from django import forms

from .models import Schedule, MessageSet, Message, BinaryContent


class CloneMessageSetForm(forms.Form):
    short_name = forms.CharField(label='New short name', required=True)

    def clean_short_name(self):
        short_name = self.cleaned_data['short_name']
        if MessageSet.objects.filter(short_name=short_name).exists():
            raise forms.ValidationError(
                'A message set already exists with this name.')
        return short_name


class MessageSetAdmin(admin.ModelAdmin):

    actions = ['clone_messageset']

    def clone_messageset(self, request, queryset):
        if queryset.count() > 1:
            self.message_user(
                request, 'Only 1 message set can be cloned at a time.',
                level=messages.WARNING)
            return

        original = queryset.first()
        if 'do_action' in request.POST:
            form = CloneMessageSetForm(request.POST)
            if form.is_valid():
                new_short_name = form.cleaned_data['short_name']

                # NOTE: setting the PK to `None` results in a new record being
                # #     created
                clone = queryset.first()
                clone.pk = None
                clone.short_name = new_short_name
                clone.save()

                for cloned_message in original.messages.all():
                    cloned_message.pk = None
                    cloned_message.messageset = clone
                    cloned_message.save()

                self.message_user(request, 'Cloned the message set')
                return
        else:
            form = CloneMessageSetForm()

        return render(
            request, 'contentstore/messageset/clone_messageset.html',
            {
                'title': u'Clone message set: %s' % (original,),
                'object': original,
                'form': form,
            })
    clone_messageset.short_description = 'Clone selected message set'


admin.site.register(Schedule)
admin.site.register(MessageSet, MessageSetAdmin)
admin.site.register(Message)
admin.site.register(BinaryContent)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.11 on 2017-01-16 07:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions', '0003_auto_20160322_1534'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscription',
            name='initial_sequence_number',
            field=models.IntegerField(default=1),
        ),
    ]



try:
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlunparse

from celery.task import Task
from celery.utils.log import get_task_logger
from celery.exceptions import SoftTimeLimitExceeded

from django.db.models import Count, Q
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.sites.shortcuts import get_current_site
from django.utils.timezone import now
from seed_services_client.metrics import MetricsApiClient
from requests import exceptions as requests_exceptions

from .models import Subscription, SubscriptionSendFailure, EstimatedSend
from seed_stage_based_messaging import utils
from contentstore.models import Message, MessageSet, Schedule
from seed_services_client import MessageSenderApiClient, SchedulerApiClient

logger = get_task_logger(__name__)


def get_metric_client(session=None):
    return MetricsApiClient(
        url=settings.METRICS_URL,
        auth=settings.METRICS_AUTH,
        session=session)


def make_absolute_url(path):
    # NOTE: We're using the default site as set by
    #       settings.SITE_ID and the Sites framework
    site = get_current_site(None)
    return urlunparse(
        ('https' if settings.USE_SSL else 'http',
         site.domain, path,
         '', '', ''))


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """
    name = "subscriptions.tasks.fire_metric"

    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {
            metric_name: metric_value
        }
        metric_client = get_metric_client(session=session)
        metric_client.fire_metrics(**metric)
        return "Fired metric <%s> with value <%s>" % (
            metric_name, metric_value)


fire_metric = FireMetric()


class SendNextMessage(Task):

    """
    Task to load and contruct message and send them off
    """
    name = "subscriptions.tasks.send_next_message"
    default_retry_delay = 5
    max_retries = 5

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def run(self, subscription_id, **kwargs):
        """
        Load and contruct message and send them off
        """
        log = self.get_logger(**kwargs)

        log.info("Loading Subscription")
        try:
            subscription = Subscription.objects.get(id=subscription_id)
        except Subscription.DoesNotExist:
            logger.error('Could not load subscription <%s>' % subscription_id,
                         exc_info=True)
            return

        if not subscription.is_ready_for_processing:
            if (subscription.process_status == 2 or
                    subscription.completed is True):
                # Disable the subscription's scheduler
                schedule_disable.apply_async(args=[subscription_id])
                log.info("Scheduler deactivation task fired")
                return "Schedule deactivation task fired"

            else:
                log.info("Message sending aborted - busy, broken or inactive")
                # TODO: retry if busy (process_status = 1)
                # TODO: be more specific about why it aborted
                return ("Message sending aborted, status <%s>" %
                        subscription.process_status)

        try:
            log.info("Loading Message")
            message = Message.objects.get(
                messageset=subscription.messageset,
                sequence_number=subscription.next_sequence_number,
                lang=subscription.lang)
        except ObjectDoesNotExist:
            error = ('Missing Message: MessageSet: <%s>, Sequence Number: <%s>'
                     ', Lang: <%s>') % (
                subscription.messageset,
                subscription.next_sequence_number,
                subscription.lang)
            logger.error(error, exc_info=True)
            return "Message sending aborted, missing message"

        if self.request.retries > 0:
            retry_delay = utils.calculate_retry_delay(self.request.retries)
        else:
            retry_delay = self.default_retry_delay

        # Start processing
        log.debug("setting process status to 1")
        subscription.process_status = 1  # in process
        log.debug("saving subscription")
        subscription.save()

        log.info("Loading Initial Recipient Identity")
        try:
            to_addr = utils.get_identity_address(
                subscription.identity,
                use_communicate_through=True)
        except requests_exceptions.ConnectionError as exc:
            log.info('Connection Error to Identity Store')
            fire_metric.delay('sbm.send_next_message.connection_error.sum', 1)
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.HTTPError as exc:
            # Recoverable HTTP errors: 500, 401
            log.info('Identity Store Request failed due to status: %s' %
                     exc.response.status_code)
            metric_name = ('sbm.send_next_message.http_error.%s.sum' %
                           exc.response.status_code)
            fire_metric.delay(metric_name, 1)
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.Timeout as exc:
            log.info('Identity Store Request failed due to timeout')
            fire_metric.delay('sbm.send_next_message.timeout.sum', 1)
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)
        log.debug("to_addr determined - %s" % to_addr)

        if to_addr is None:
            log.info("No valid recipient to_addr found")
            subscription.process_status = -1  # Error
            log.debug("saving subscription")
            subscription.save()
            log.debug("Firing error metric")
            fire_metric.apply_async(kwargs={
                "metric_name": 'subscriptions.send_next_message_errored.sum',  # noqa
                "metric_value": 1.0
            })
            log.debug("Fired error metric")
            return "Valid recipient could not be found"

        # All preconditions have been met
        log.info("Preparing message payload with: %s" % message.id)  # noqa
        payload = {
            "to_addr": to_addr,
            "to_identity": subscription.identity,
            "delivered": "false",
            "metadata": {}
        }

        if subscription.messageset.channel:
            payload["channel"] = subscription.messageset.channel

        prepend_next = None
        if subscription.messageset.content_type == "text":
            log.debug("Determining payload content")
            if subscription.metadata is not None and \
               "prepend_next_delivery" in subscription.metadata \
               and subscription.metadata["prepend_next_delivery"] is not None:  # noqa
                prepend_next = subscription.metadata["prepend_next_delivery"]
                log.debug("Prepending next delivery")
                payload["content"] = "%s\n%s" % (
                    subscription.metadata["prepend_next_delivery"],
                    message.text_content)
                # clear prepend_next_delivery
                log.debug("Clearing prepended message")
                subscription.metadata[
                    "prepend_next_delivery"] = None
                subscription.save()
            else:
                log.debug("Loading default content")
                payload["content"] = message.text_content
            log.debug("text content loaded")
        else:
            # TODO - audio media handling on MC
            # audio

            if subscription.metadata is not None and \
               "prepend_next_delivery" in subscription.metadata \
               and subscription.metadata["prepend_next_delivery"] is not None:  # noqa
                prepend_next = subscription.metadata["prepend_next_delivery"]
                payload["metadata"]["voice_speech_url"] = [
                    subscription.metadata["prepend_next_delivery"],
                    make_absolute_url(
                        message.binary_content.content.url),
                ]
                # clear prepend_next_delivery
                subscription.metadata[
                    "prepend_next_delivery"] = None
                subscription.save()
            else:
                payload["metadata"]["voice_speech_url"] = \
                    make_absolute_url(
                        message.binary_content.content.url)

        log.info("Sending message to Message Sender")
        message_sender_client = MessageSenderApiClient(
            settings.MESSAGE_SENDER_TOKEN,
            settings.MESSAGE_SENDER_URL,
            retries=5,
            timeout=settings.DEFAULT_REQUEST_TIMEOUT,
        )
        try:
            result = message_sender_client.create_outbound(payload)
        except requests_exceptions.ConnectionError as exc:
            log.info('Connection Error to Message Sender')
            fire_metric.delay('sbm.send_next_message.connection_error.sum', 1)
            # Reset the prepend next delivery that was cleared above.
            if prepend_next is not None:
                subscription.metadata["prepend_next_delivery"] = prepend_next
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.HTTPError as exc:
            # Recoverable HTTP errors: 500, 401
            log.info('Message Sender Request failed due to status: %s' %
                     exc.response.status_code)
            metric_name = ('sbm.send_next_message.http_error.%s.sum' %
                           exc.response.status_code)
            fire_metric.delay(metric_name, 1)
            # Reset the prepend next delivery that was cleared above.
            if prepend_next is not None:
                subscription.metadata["prepend_next_delivery"] = prepend_next
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)
        except requests_exceptions.Timeout as exc:
            log.info('Message Sender Request failed due to timeout')
            fire_metric.delay('sbm.send_next_message.timeout.sum', 1)
            # Reset the prepend next delivery that was cleared above.
            if prepend_next is not None:
                subscription.metadata["prepend_next_delivery"] = prepend_next
            subscription.process_status = 0
            subscription.save()
            self.retry(exc=exc, countdown=retry_delay)

        log.debug("setting process status back to 0")
        subscription.process_status = 0  # ready
        log.debug("saving subscription")
        subscription.save()

        log.debug("starting post_send_process task")
        post_send_process(subscription_id)
        log.debug("finished post_send_process task")

        log.debug("Firing SMS/OBD calls sent per message set metric")
        send_type = utils.normalise_metric_name(
                        subscription.messageset.content_type)
        ms_name = utils.normalise_metric_name(
                        subscription.messageset.short_name)
        fire_metric.apply_async(kwargs={
            "metric_name":
                'message.{}.{}.sum'.format(send_type, ms_name),
            "metric_value": 1.0
        })
        fire_metric.apply_async(kwargs={
            "metric_name":
                'message.{}.sum'.format(send_type),
            "metric_value": 1.0
        })

        log.debug("Message queued for send. ID: <%s>" % str(result["id"]))  # noqa
        return "Message queued for send. ID: <%s>" % str(result["id"])  # noqa

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        if self.request.retries == self.max_retries:
            SubscriptionSendFailure.objects.create(
                subscription_id=args[0],
                initiated_at=self.request.eta,
                reason=einfo.exception.message,
                task_id=task_id
            )
        super(SendNextMessage, self).on_failure(exc, task_id, args,
                                                kwargs, einfo)


send_next_message = SendNextMessage()


class PostSendProcess(Task):

    """
    Task to ensure subscription is bumped or converted
    """
    name = "subscriptions.tasks.post_send_process"

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def run(self, subscription_id, **kwargs):
        """
        Load subscription and process
        """
        log = self.get_logger(**kwargs)

        log.info("Loading Subscription")
        # Process moving to next message, next set or finished
        try:
            subscription = Subscription.objects.get(id=subscription_id)
            if subscription.process_status == 0:
                log.debug("setting process status to 1")
                subscription.process_status = 1  # in process
                log.debug("saving subscription")
                subscription.save()
                # Get set max
                set_max = subscription.messageset.messages.filter(
                    lang=subscription.lang).count()
                log.debug("set_max calculated - %s" % set_max)
                # Compare user position to max
                if subscription.next_sequence_number == set_max:
                    # Mark current as completed
                    log.debug("setting subscription completed")
                    subscription.completed = True
                    log.debug("setting subscription inactive")
                    subscription.active = False
                    log.debug("setting process status to 2")
                    subscription.process_status = 2  # Completed
                    log.debug("saving subscription")
                    subscription.save()
                    # If next set defined create new subscription
                    messageset = subscription.messageset
                    if messageset.next_set:
                        log.info("Creating new subscription for next set")
                        newsub = Subscription.objects.create(
                            identity=subscription.identity,
                            lang=subscription.lang,
                            messageset=messageset.next_set,
                            schedule=messageset.next_set.default_schedule
                        )
                        log.debug("Created Subscription <%s>" % newsub.id)
                else:
                    # More in this set so interate by one
                    log.debug("incrementing next_sequence_number")
                    subscription.next_sequence_number += 1
                    log.debug("setting process status back to 0")
                    subscription.process_status = 0
                    log.debug("saving subscription")
                    subscription.save()
                # return response
                return "Subscription for %s updated" % str(
                    subscription.id)
            else:
                log.info("post_send_process not executed")
                return "post_send_process not executed"

        except ObjectDoesNotExist:
            log.debug("subscription errored")
            subscription.process_status = -1  # Errored
            log.debug("saving subscription")
            subscription.save()
            logger.error('Unexpected error', exc_info=True)

        except SoftTimeLimitExceeded:
            logger.error(
                'Soft time limit exceed processing message send search '
                'via Celery.',
                exc_info=True)

        return False


post_send_process = PostSendProcess()


class ScheduleDisable(Task):

    """ Task to disable a subscription's schedule
    """
    name = "subscriptions.tasks.schedule_disable"

    def scheduler_client(self):
        return SchedulerApiClient(
            settings.SCHEDULER_API_TOKEN,
            settings.SCHEDULER_URL)

    def run(self, subscription_id, **kwargs):
        log = self.get_logger(**kwargs)
        log.info("Disabling schedule for <%s>" % (subscription_id,))
        try:
            subscription = Subscription.objects.get(id=subscription_id)
            try:
                schedule_id = subscription.metadata["scheduler_schedule_id"]
                scheduler = self.scheduler_client()
                scheduler.update_schedule(
                    subscription.metadata["scheduler_schedule_id"],
                    {"enabled": False}
                )
                log.info("Disabled schedule <%s> on scheduler for sub <%s>" % (
                    schedule_id, subscription_id))
                return True
            except Exception:
                log.info("Schedule id not saved in subscription metadata")
                return False
        except ObjectDoesNotExist:
            logger.error('Missing Subscription', exc_info=True)
        except SoftTimeLimitExceeded:
            logger.error(
                'Soft time limit exceed processing schedule create '
                'via Celery.',
                exc_info=True)
        return False


schedule_disable = ScheduleDisable()


class ScheduleCreate(Task):

    """ Task to tell scheduler a new subscription created
    """
    name = "subscriptions.tasks.schedule_create"

    def scheduler_client(self):
        return SchedulerApiClient(
            settings.SCHEDULER_API_TOKEN,
            settings.SCHEDULER_URL)

    def run(self, subscription_id, **kwargs):
        """ Returns remote scheduler_id UUID
        """

        log = self.get_logger(**kwargs)
        log.info("Creating schedule for <%s>" % (subscription_id,))
        try:
            subscription = Subscription.objects.get(id=subscription_id)
            if subscription.process_status == 0:
                schedule = {
                    "frequency": None,
                    "cron_definition": subscription.schedule.cron_string,
                    "endpoint": "%s/%s/send" % (
                        settings.STAGE_BASED_MESSAGING_URL, subscription_id),
                    "auth_token": settings.SCHEDULER_INBOUND_API_TOKEN
                }
                scheduler = self.scheduler_client()
                result = scheduler.create_schedule(schedule)
                log.info("Created schedule <%s> on scheduler for sub <%s>" % (
                    result["id"], subscription_id))
                if subscription.metadata is None:
                    subscription.metadata = {"scheduler_schedule_id": result["id"]}  # noqa
                else:
                    subscription.metadata["scheduler_schedule_id"] = result["id"]  # noqa
                subscription.save()
                return result["id"]

        except ObjectDoesNotExist:
            logger.error('Missing Subscription', exc_info=True)

        except SoftTimeLimitExceeded:
            logger.error(
                'Soft time limit exceed processing schedule create '
                'via Celery.',
                exc_info=True)

        return False


schedule_create = ScheduleCreate()


class ScheduledMetrics(Task):

    """ Fires off tasks for all the metrics that should run
        on a schedule
    """
    name = "subscriptions.tasks.scheduled_metrics"

    def run(self, **kwargs):
        globs = globals()  # execute globals() outside for loop for efficiency
        for metric in settings.METRICS_SCHEDULED_TASKS:
            globs[metric].apply_async()

        return "%d Scheduled metrics launched" % len(
            settings.METRICS_SCHEDULED_TASKS)


scheduled_metrics = ScheduledMetrics()


class FireActiveLast(Task):

    """ Fires last active subscriptions count
    """
    name = "subscriptions.tasks.fire_active_last"

    def run(self):
        active_subs = Subscription.objects.filter(active=True).count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.active.last',
            "metric_value": active_subs
        })


fire_active_last = FireActiveLast()


class FireCreatedLast(Task):

    """ Fires last created subscriptions count
    """
    name = "subscriptions.tasks.fire_created_last"

    def run(self):
        created_subs = Subscription.objects.all().count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.created.last',
            "metric_value": created_subs
        })


fire_created_last = FireCreatedLast()


class FireBrokenLast(Task):

    """ Fires last broken subscriptions count
    """
    name = "subscriptions.tasks.fire_broken_last"

    def run(self):
        broken_subs = Subscription.objects.filter(process_status=-1).count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.broken.last',
            "metric_value": broken_subs
        })


fire_broken_last = FireBrokenLast()


class FireCompletedLast(Task):

    """ Fires last completed subscriptions count
    """
    name = "subscriptions.tasks.fire_completed_last"  # noqa

    def run(self):
        completed_subs = Subscription.objects.filter(completed=True).count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.completed.last',
            "metric_value": completed_subs
        })


fire_completed_last = FireCompletedLast()


class FireIncompleteLast(Task):

    """ Fires last incomplete subscriptions count
    """
    name = "seed_stage_based_messaging.subscriptions.tasks.fire_incomplete_last"  # noqa

    def run(self):
        incomplete_subs = Subscription.objects.filter(completed=False).count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.incomplete.last',
            "metric_value": incomplete_subs
        })


fire_incomplete_last = FireIncompleteLast()


class FireMessageSetsTasks(Task):

    """ Fires off seperate tasks to count active subscriptions for each
        messageset found
    """
    name = "subscriptions.tasks.fire_messagesets_tasks"  # noqa

    def run(self, **kwargs):
        # get message sets
        messagesets = MessageSet.objects.all()
        for messageset in messagesets:
            fire_messageset_last.apply_async(kwargs={
                "msgset_id": messageset.id,
                "short_name": messageset.short_name
            })
        return "%d MessageSet metrics launched" % messagesets.count()


fire_messagesets_tasks = FireMessageSetsTasks()


class FireMessageSetLast(Task):

    name = "subscriptions.tasks.fire_messageset_last"  # noqa

    def run(self, msgset_id, short_name, **kwargs):
        active_msgset_subs = Subscription.objects.filter(
            messageset=msgset_id, active=True).count()
        return fire_metric.apply_async(kwargs={
            "metric_name": 'subscriptions.%s.active.last' % short_name,
            "metric_value": active_msgset_subs
        })


fire_messageset_last = FireMessageSetLast()


class FireWeekEstimateLast(Task):
    """Fires week estimated send counts.
    """
    name = "subscriptions.tasks.fire_week_estimate_last"

    def run(self):
        schedules = Schedule.objects.filter(
            subscriptions__active=True,
            subscriptions__completed=False,
            subscriptions__process_status=0
        ).annotate(total_subs=Count('subscriptions'))
        totals = {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0}
        for schedule in schedules:
            for day in range(7):
                if (str(day) in schedule.day_of_week or
                        '*' in schedule.day_of_week):
                    totals[day] = totals[day] + schedule.total_subs

        # Django's datetime's weekday method has Monday = 0
        # whereas the cron format used in the schedules has Sunday = 0
        sunday = totals.pop(0)
        totals[7] = sunday
        totals = {(k-1): v for k, v in totals.items()}

        today = now()
        for dow, total in totals.items():
            # Only fire the metric for today or days in the future so that
            # estimates for the week don't get updated after the day in
            # question.
            if dow >= (today.weekday()):
                fire_metric.apply_async(kwargs={
                    "metric_name": 'subscriptions.send.estimate.%s.last' % dow,
                    "metric_value": total
                })


fire_week_estimate_last = FireWeekEstimateLast()


class FireDailySendEstimate(Task):
    """Fires daily estimated send counts.
    """
    name = "subscriptions.tasks.fire_daily_send_estimate"

    def run(self):
        # Django's datetime's weekday method has Monday = 0
        # whereas the cron format used in the schedules has Sunday = 0
        day = now().weekday() + 1

        schedules = Schedule.objects.filter(
            Q(day_of_week__contains=day) | Q(day_of_week__contains='*'),
            subscriptions__active=True,
            subscriptions__completed=False,
            subscriptions__process_status=0
        ).values('subscriptions__messageset').annotate(
            total_subs=Count('subscriptions'),
            total_unique=Count('subscriptions__identity', distinct=True))

        for schedule in schedules:
            EstimatedSend.objects.get_or_create(
                send_date=now().date(),
                messageset_id=schedule['subscriptions__messageset'],
                estimate_subscriptions=schedule['total_subs'],
                estimate_identities=schedule['total_unique']
            )


fire_daily_send_estimate = FireDailySendEstimate()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed schedules.
    """
    name = "subscriptions.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        log = self.get_logger(**kwargs)
        failures = SubscriptionSendFailure.objects
        log.info("Attempting to requeue <%s> failed Subscription sends" %
                 failures.all().count())
        for failure in failures.iterator():
            subscription_id = str(failure.subscription_id)
            # Cleanup the failure before requeueing it.
            failure.delete()
            send_next_message.delay(subscription_id)


requeue_failed_tasks = RequeueFailedTasks()



# Enable referencing Django settings with intersphinx
def setup(app):
    app.add_crossref_type(
        directivename='setting',
        rolename='setting',
        indextemplate='pair: %s; setting',
    )



from django.core.management.base import BaseCommand
from django.core.validators import URLValidator
from django.conf import settings
from requests import exceptions
from demands import HTTPServiceError

from subscriptions.models import Subscription

from seed_services_client import HubApiClient, IdentityStoreApiClient


def url_validator(url_str):
    URLValidator(url_str)
    return url_str


class Command(BaseCommand):
    help = ("Searches all subscription with status 5 and updates the linked "
            "identity and registration object. This is for excluding the "
            "relevant items from reports")

    def add_arguments(self, parser):
        parser.add_argument(
            '--hub-url',
            type=url_validator,
            help='The HUB url to use'
        )
        parser.add_argument(
            '--hub-token',
            type=str,
            help=('The HUB API token')
        )
        parser.add_argument(
            '--identity-url',
            type=url_validator,
            default=settings.IDENTITY_STORE_URL,
            help='The Identity Store url to use'
        )
        parser.add_argument(
            '--identity-token',
            type=str,
            default=settings.IDENTITY_STORE_TOKEN,
            help=('The Identity Store API token')
        )
        parser.add_argument(
            '--hub-identity-field',
            type=str,
            default='mother_id',
            help=('The HUB Identity field')
        )

    def handle(self, *args, **options):
        hub_url = options['hub_url']
        hub_token = options['hub_token']
        id_url = options['identity_url']
        id_token = options['identity_token']
        hub_id_field = options['hub_identity_field']

        if not hub_url or not hub_token:
            self.warning("hub-url and hub-token is required.")
            return

        hubApi = HubApiClient(hub_token, hub_url)
        idApi = IdentityStoreApiClient(id_token, id_url)

        subscriptions = Subscription.objects.filter(process_status=5)

        reg_count = 0
        id_count = 0
        for subscription in subscriptions.iterator():
            identity = subscription.identity

            # find and update linked registrations
            registrations = hubApi.get_registrations({hub_id_field: identity})

            for registration in registrations['results']:

                if not registration['data'].get('exclude_report', False):
                    registration['data']['exclude_report'] = True

                    try:
                        hubApi.update_registration(
                            registration['id'], {'data': registration['data']})
                        reg_count += 1
                    except exceptions.ConnectionError as exc:
                        self.warning(
                            'Connection error to Hub API: {}'.format(exc))
                    except HTTPServiceError as exc:
                        self.warning('Invalid Hub API response({}): {}'.format(
                            exc.response.status_code, exc.response.url))

            # find and update linked identities
            identity = idApi.get_identity(identity)

            if not identity['details'].get('exclude_report', False):
                identity['details']['exclude_report'] = True

                try:
                    idApi.update_identity(identity['id'],
                                          {'details': identity['details']})
                    id_count += 1
                except exceptions.ConnectionError as exc:
                    self.warning(
                        'Connection error to Identity API: {}'.format(exc))
                except HTTPServiceError as exc:
                    self.warning(
                        'Invalid Identity Store API response({}): {}'.format(
                            exc.response.status_code, exc.response.url))

        self.success('Updated %s identities and %s registrations.' % (
            id_count, reg_count))

    def log(self, level, msg):
        self.stdout.write(level(msg))

    def warning(self, msg):
        self.log(self.style.WARNING, msg)

    def success(self, msg):
        self.log(self.style.SUCCESS, msg)



import random
import re
from django.conf import settings
from contentstore.models import MessageSet
from subscriptions.models import Subscription
from seed_services_client import IdentityStoreApiClient

NORMALISE_METRIC_RE = re.compile(r'\W+')

identity_store_client = IdentityStoreApiClient(
    settings.IDENTITY_STORE_TOKEN,
    settings.IDENTITY_STORE_URL,
    retries=5,
    timeout=settings.DEFAULT_REQUEST_TIMEOUT,
)


def get_identity(identity_uuid):
    return identity_store_client.get_identity(identity_uuid)


def normalise_metric_name(name):
    """
    Replaces all non-alphanumeric characters with underscores.
    """
    return NORMALISE_METRIC_RE.sub('_', name).rstrip('_').lstrip('_')


def get_identity_address(identity_uuid, use_communicate_through=False):
    params = {"default": True}
    if use_communicate_through:
        params['use_communicate_through'] = True

    return identity_store_client.get_identity_address(
        identity_uuid, params=params)


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    messagesets = MessageSet.objects.all()
    for messageset in messagesets:
        messageset_name = normalise_metric_name(messageset.short_name)
        available_metrics.append(
            "subscriptions.%s.active.last" % messageset.short_name)
        available_metrics.append(
            "subscriptions.message_set.{}.sum".format(messageset_name))
        available_metrics.append(
            "subscriptions.message_set.{}.total.last".format(messageset_name))

    languages = Subscription.objects.order_by('lang').distinct('lang')\
        .values_list('lang', flat=True)
    for lang in languages:
        lang_normal = normalise_metric_name(lang)
        available_metrics.append(
            "subscriptions.language.{}.sum".format(lang_normal))
        available_metrics.append(
            "subscriptions.language.{}.total.last".format(lang_normal))

    content_types = MessageSet._meta.get_field('content_type').choices
    for content_type in content_types:
        type_normal = normalise_metric_name(content_type[0])
        available_metrics.append(
            "subscriptions.message_format.{}.sum".format(type_normal))
        available_metrics.append(
            "subscriptions.message_format.{}.total.last".format(type_normal))
        available_metrics.append(
            "message.{}.sum".format(type_normal))

        for messageset in messagesets:
            messageset_name = normalise_metric_name(messageset.short_name)
            available_metrics.append(
                "message.{}.{}.sum".format(type_normal, messageset_name))

    return available_metrics


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-09 13:14
from __future__ import unicode_literals

import contentstore.models
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BinaryContent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content', models.FileField(upload_to=contentstore.models.generate_new_filename)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sequence_number', models.IntegerField()),
                ('lang', models.CharField(max_length=6)),
                ('text_content', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('binary_content', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='message', to='contentstore.BinaryContent')),
            ],
            options={
                'ordering': ['sequence_number'],
            },
        ),
        migrations.CreateModel(
            name='MessageSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('short_name', models.CharField(max_length=20, unique=True, verbose_name='Short name')),
                ('notes', models.TextField(blank=True, null=True, verbose_name='Notes')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='Schedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('minute', models.CharField(default='*', max_length=64, verbose_name='minute')),
                ('hour', models.CharField(default='*', max_length=64, verbose_name='hour')),
                ('day_of_week', models.CharField(default='*', max_length=64, verbose_name='day of week')),
                ('day_of_month', models.CharField(default='*', max_length=64, verbose_name='day of month')),
                ('month_of_year', models.CharField(default='*', max_length=64, verbose_name='month of year')),
            ],
            options={
                'verbose_name_plural': 'schedules',
                'ordering': ['month_of_year', 'day_of_month', 'day_of_week', 'hour', 'minute'],
                'verbose_name': 'schedule',
            },
        ),
        migrations.AddField(
            model_name='messageset',
            name='default_schedule',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='message_sets', to='contentstore.Schedule'),
        ),
        migrations.AddField(
            model_name='messageset',
            name='next_set',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contentstore.MessageSet'),
        ),
        migrations.AddField(
            model_name='message',
            name='messageset',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='contentstore.MessageSet'),
        ),
    ]



from seed_stage_based_messaging.settings import *  # flake8: noqa

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'TESTSEKRET'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
CELERY_ALWAYS_EAGER = True
BROKER_BACKEND = 'memory'
CELERY_RESULT_BACKEND = 'djcelery.backends.database:DatabaseBackend'

SCHEDULER_URL = "http://seed-scheduler/api/v1"
SCHEDULER_API_TOKEN = "REPLACEME"

IDENTITY_STORE_URL = "http://seed-identity-store/api/v1"
IDENTITY_STORE_TOKEN = "REPLACEME"

MESSAGE_SENDER_URL = "http://seed-message-sender/api/v1"
MESSAGE_SENDER_TOKEN = "REPLACEME"

METRICS_URL = "http://metrics-url"
METRICS_AUTH_TOKEN = "REPLACEME"

PASSWORD_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher',)


# REST Framework conf defaults
REST_FRAMEWORK['PAGE_SIZE'] = 2



# -*- coding: utf-8 -*-

from django.test import TestCase
from rest_framework.serializers import ValidationError

from ..models import Message, validate_special_characters
from .test_general import MessageSetTestMixin


class TestValidators(TestCase):
    def test_special_characters_allows_normal_text(self):
        self.assertEqual(validate_special_characters('ascii text!'), None)

    def test_special_characters_rejects_bad_chars(self):
        with self.assertRaises(ValidationError):
            validate_special_characters(u'I said “hello"')

        with self.assertRaises(ValidationError):
            validate_special_characters(u'I said "hello”')

        with self.assertRaises(ValidationError):
            validate_special_characters(u'It’s over there')

        with self.assertRaises(ValidationError):
            validate_special_characters(u'Hello –')


class TestMessage(MessageSetTestMixin, TestCase):
    def test_raises_validation_error_for_special_chars(self):
        messageset = self.make_messageset()
        special_chars = u'This text has a special character – there'

        with self.assertRaises(ValidationError):
            m = Message.objects.create(
                sequence_number=1,
                messageset_id=messageset.id,
                text_content=special_chars,)
            m.full_clean()



from django.conf.urls import url, include
from rest_framework import routers
from . import views

router = routers.DefaultRouter()
router.register(r'subscriptions', views.SubscriptionViewSet)
router.register(r'failed-tasks', views.FailedTaskViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url(r'^api/v1/', include(router.urls)),
    url(r'^api/v1/subscriptions/(?P<subscription_id>.+)/send$',
        views.SubscriptionSend.as_view()),
    url(r'^api/v1/subscriptions/request$',
        views.SubscriptionRequest.as_view()),
    url(r'^api/v1/user/token/$', views.UserView.as_view(),
        name='create-user-token'),
    url(r'^api/v1/runsendestimate$',
        views.DailyEstimateRun.as_view()),
]



from rest_framework import filters, viewsets, status, mixins
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.authtoken.models import Token
from django.contrib.auth.models import User
import django_filters

from .models import Subscription, SubscriptionSendFailure
from .serializers import (SubscriptionSerializer, CreateUserSerializer,
                          SubscriptionSendFailureSerializer)
from .tasks import (send_next_message, scheduled_metrics, requeue_failed_tasks,
                    fire_daily_send_estimate)
from seed_stage_based_messaging.utils import get_available_metrics


class CreatedAtCursorPagination(CursorPagination):
    ordering = "-created_at"


class IdCursorPagination(CursorPagination):
    ordering = "-id"


class SubscriptionFilter(filters.FilterSet):
    created_after = django_filters.IsoDateTimeFilter(
        name="created_at", lookup_type="gte")
    created_before = django_filters.IsoDateTimeFilter(
        name="created_at", lookup_type="lte")

    class Meta:
        model = Subscription


class SubscriptionViewSet(viewsets.ModelViewSet):

    """ API endpoint that allows Subscription models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Subscription.objects.all()
    serializer_class = SubscriptionSerializer
    filter_class = SubscriptionFilter
    pagination_class = CreatedAtCursorPagination


class SubscriptionSend(APIView):

    """ Triggers a send for the next subscription message
    """
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        """ Validates subscription data before creating Outbound message
        """
        # Look up subscriber
        subscription_id = kwargs["subscription_id"]
        if Subscription.objects.filter(id=subscription_id).exists():
            status = 201
            accepted = {"accepted": True}
            send_next_message.apply_async(args=[subscription_id])
        else:
            status = 400
            accepted = {"accepted": False,
                        "reason": "Missing subscription in control"}
        return Response(accepted, status=status)


class SubscriptionRequest(APIView):

    """ Webhook listener for registrations now needing a subscription
    """
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        """ Validates subscription data before creating Subscription message
        """
        # Ensure that we check for the 'data' key in the request object before
        # attempting to reference it
        if "data" in request.data:
            # This is a workaround for JSONField not liking blank/null refs
            if "metadata" not in request.data["data"]:
                request.data["data"]["metadata"] = {}

            if "initial_sequence_number" not in request.data["data"]:
                request.data["data"]["initial_sequence_number"] = \
                    request.data["data"].get("next_sequence_number")

            subscription = SubscriptionSerializer(data=request.data["data"])
            if subscription.is_valid():
                subscription.save()
                # Return
                status = 201
                accepted = {"accepted": True}
                return Response(accepted, status=status)
            else:
                status = 400
                return Response(subscription.errors, status=status)
        else:
            status = 400
            message = {"data": ["This field is required."]}
            return Response(message, status=status)


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """
    permission_classes = (IsAdminUser,)

    def post(self, request):
        '''Create a user and token, given an email. If user exists just
        provide the token.'''
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get('email')
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(
            status=status.HTTP_201_CREATED, data={'token': token.key})


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {
            "metrics_available": get_available_metrics()
        }
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_stage_based_messaging
        import django
        import rest_framework
        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_stage_based_messaging.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__
                }
            }
        }
        return Response(resp, status=status)


class FailedTaskViewSet(mixins.ListModelMixin,
                        mixins.RetrieveModelMixin,
                        viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = SubscriptionSendFailure.objects.all()
    serializer_class = SubscriptionSendFailureSerializer
    pagination_class = IdCursorPagination

    def create(self, request):
        status = 201
        resp = {'requeued_failed_tasks': True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)


class DailyEstimateRun(APIView):

    """ Triggers a daily send estimation
    """
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        task_id = fire_daily_send_estimate.apply_async()
        accepted = {"accepted": True, "task_id": str(task_id)}
        return Response(accepted, status=202)



from django.db.models import Count
from django.core import exceptions
from django.core.management.base import BaseCommand
from django.utils.six.moves import input

from subscriptions.models import Subscription
from contentstore.models import MessageSet


def list_validator(valid_inputs):
    def validate(value):
        if value in valid_inputs:
            return value
        else:
            msg = "Selelection must be one of the following: {0}".format(
                str(valid_inputs))
            raise exceptions.ValidationError(msg, code='invalid_choice')

    return validate


class Command(BaseCommand):
    help = "Resets subscriptions status to ready."

    def get_option(self, value, option_format='[{0:>3}]'):
        return option_format.format(value)

    def handle(self, *args, **options):
        selection = None
        confirm = None

        stuck_subscriptions = Subscription.objects.filter(process_status=1)
        stuck_totals = stuck_subscriptions.values('messageset')\
            .annotate(total=Count('messageset')).order_by('messageset')

        self.stdout.write("Totals of stuck Subscriptions per MessageSet:")
        for total in stuck_totals:
            ms = MessageSet.objects.get(pk=total['messageset'])
            opt = self.get_option(ms.pk)
            line = "{opt} {name:<40} Total: {total:>6}".format(
                opt=opt,
                name=ms.short_name,
                total=total['total'])
            self.stdout.write(line)

        line = "{opt} {name:<40} Total: {total:>6}".format(
            opt=self.get_option('all'),
            name="All MessageSets",
            total=stuck_subscriptions.count())
        self.stdout.write(line)
        line = "{opt} Do nothing and quit".format(
            opt=self.get_option('q'),
            name="All MessageSets",
            total=stuck_subscriptions.count())
        self.stdout.write(line)

        msg = ("Please type the ID of the MessageSet you'd like to reset "
               "subscriptions for: ")
        valid_inputs = [str(o['messageset']) for o in stuck_totals]
        valid_inputs.extend(['all', 'q'])
        while selection is None:
            selection = self.get_input_data(msg, list_validator(valid_inputs))
            if selection == 'q':
                self.stdout.write("Quiting.")
                return

        if selection == 'all':
            records = stuck_subscriptions.all()
        else:
            records = stuck_subscriptions.filter(messageset=selection)

        msg = ("Do you wish to update these subscriptions to "
               "ready status? [y/n] ")
        while confirm is None:
            confirm = self.get_input_data(msg, list_validator(['y', 'n']))

        if confirm:
            rows_updated = records.update(process_status=0)
            msg = "Updated {0} rows to ready status".format(rows_updated)
            self.stdout.write(self.style.SUCCESS(msg))
        else:
            self.stdout.write("Ok, doing nothing.")

    def get_input_data(self, message, validator=None):
        raw_value = input(message)
        if validator is not None:
            try:
                val = validator(raw_value)
            except exceptions.ValidationError as e:
                self.stderr.write("Error: %s" % '; '.join(e.messages))
                val = None
        return val



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-13 09:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions', '0004_subscription_initial_sequence_number'),
    ]

    operations = [
        migrations.CreateModel(
            name='SubscriptionSendFailure',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.UUIDField()),
                ('initiated_at', models.DateTimeField()),
                ('reason', models.TextField()),
                ('subscription', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='subscriptions.Subscription')),
            ],
        ),
    ]



from .models import Schedule, MessageSet, Message, BinaryContent

from rest_framework import serializers


class ScheduleSerializer(serializers.ModelSerializer):

    class Meta:
        model = Schedule
        fields = ('id', 'minute', 'hour', 'day_of_week', 'day_of_month',
                  'month_of_year')


class MessageSetSerializer(serializers.ModelSerializer):

    class Meta:
        model = MessageSet
        fields = ('id', 'short_name', 'content_type', 'notes', 'next_set',
                  'default_schedule', 'created_at', 'updated_at')


class BinaryContentSerializer(serializers.ModelSerializer):

    class Meta:
        model = BinaryContent
        fields = ('id', 'content')


class MessageSerializer(serializers.ModelSerializer):

    class Meta:
        model = Message
        fields = ('id', 'messageset', 'sequence_number', 'lang',
                  'text_content', 'binary_content', 'created_at', 'updated_at')


class MessageListSerializer(serializers.ModelSerializer):

    """
        Only used for get views, because binary relational serializer is not
        something that works nicely with posts
    """
    binary_content = BinaryContentSerializer(many=False, read_only=True)

    class Meta:
        model = Message
        fields = ('id', 'messageset', 'sequence_number', 'lang',
                  'text_content', 'binary_content', 'created_at', 'updated_at')


class MessageSetMessagesSerializer(serializers.ModelSerializer):
    messages = MessageListSerializer(many=True, read_only=True)

    class Meta:
        model = MessageSet
        fields = ('id', 'short_name', 'notes', 'next_set', 'default_schedule',
                  'messages', 'created_at', 'updated_at')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-07-27 08:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('subscriptions', '0005_subscriptionsendfailure'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subscription',
            name='identity',
            field=models.CharField(db_index=True, max_length=36),
        ),
    ]



"""
Django settings for seed_stage_based_messaging project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

import os
import djcelery
import dj_database_url
import django_cache_url
import mimetypes

from kombu import Exchange, Queue

# Support SVG on admin
mimetypes.add_type("image/svg+xml", ".svg", True)
mimetypes.add_type("image/svg+xml", ".svgz", True)

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('SECRET_KEY', 'REPLACEME')

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get('DEBUG', False)

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = (
    # admin
    'django.contrib.admin',
    # core
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    # documentation
    'rest_framework_docs',
    # 3rd party
    'djcelery',
    'raven.contrib.django.raven_compat',
    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',
    # us
    'contentstore',
    'subscriptions',

)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

SITE_ID = 1
USE_SSL = os.environ.get('USE_SSL', 'false').lower() == 'true'

ROOT_URLCONF = 'seed_stage_based_messaging.urls'

WSGI_APPLICATION = 'seed_stage_based_messaging.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    'default': dj_database_url.config(
        default=os.environ.get(
            'STAGE_BASED_MESSAGING_DATABASE',
            'postgres://postgres:@localhost/seed_stage_based_messaging')),
}


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-gb'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'django.contrib.staticfiles.finders.FileSystemFinder',
)

STATIC_ROOT = 'staticfiles'
STATIC_URL = '/static/'

MEDIA_ROOT = 'mediafiles'
MEDIA_URL = '/media/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates')
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Sentry configuration
STAGE_BASED_MESSAGING_SENTRY_DSN = os.environ.get(
    'STAGE_BASED_MESSAGING_SENTRY_DSN', None)
RAVEN_CONFIG = {
    # DevOps will supply you with this.
    'dsn': STAGE_BASED_MESSAGING_SENTRY_DSN,
}

# REST Framework conf defaults
REST_FRAMEWORK = {
    'PAGE_SIZE': 1000,
    'DEFAULT_PAGINATION_CLASS':
        'rest_framework.pagination.CursorPagination',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',)
}

# Celery configuration options
CELERY_RESULT_BACKEND = 'djcelery.backends.database:DatabaseBackend'
CELERYBEAT_SCHEDULER = 'djcelery.schedulers.DatabaseScheduler'

BROKER_URL = os.environ.get(
    'BROKER_URL',
    'amqp://localhost:5672//seed_stage_based_messaging')

CELERY_DEFAULT_QUEUE = 'seed_stage_based_messaging'
CELERY_QUEUES = (
    Queue('seed_stage_based_messaging',
          Exchange('seed_stage_based_messaging'),
          routing_key='seed_stage_based_messaging'),
)

CELERY_ALWAYS_EAGER = False

# Tell Celery where to find the tasks
CELERY_IMPORTS = (
    'subscriptions.tasks',
)

CELERY_CREATE_MISSING_QUEUES = True
CELERY_ROUTES = {
    'celery.backend_cleanup': {
        'queue': 'mediumpriority',
    },
    'subscriptions.tasks.send_next_message': {
        'queue': 'priority',
    },
    'subscriptions.tasks.post_send_process': {
        'queue': 'mediumpriority',
    },
    'subscriptions.tasks.schedule_create': {
        'queue': 'mediumpriority',
    },
    'subscriptions.tasks.schedule_disable': {
        'queue': 'mediumpriority',
    },
    'subscriptions.tasks.requeue_failed_tasks': {
        'queue': 'mediumpriority',
    },
    'subscriptions.tasks.fire_metric': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.scheduled_metrics': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.fire_active_last': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.fire_created_last': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.fire_broken_last': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.fire_completed_last': {
        'queue': 'metrics',
    },
    'subscriptions.tasks.fire_incomplete_last': {
        'queue': 'metrics',
    },
}

CELERYD_MAX_TASKS_PER_CHILD = 50

CACHE_URL = os.environ.get('STAGE_BASED_MESSAGING_CACHE', 'locmem://')
CACHES = {
    'default': django_cache_url.parse(CACHE_URL),
}

METRICS_REALTIME = [
    'subscriptions.created.sum',
    'subscriptions.send_next_message_errored.sum'
]
# Note metrics with variable names of messageset short_names not included here
METRICS_SCHEDULED = [
    'subscriptions.active.last',
    'subscriptions.created.last',
    'subscriptions.broken.last',
    'subscriptions.completed.last',
    'subscriptions.send.estimate.0.last',
    'subscriptions.send.estimate.1.last',
    'subscriptions.send.estimate.2.last',
    'subscriptions.send.estimate.3.last',
    'subscriptions.send.estimate.4.last',
    'subscriptions.send.estimate.5.last',
    'subscriptions.send.estimate.6.last',
    'subscriptions.send_next_message.connection_error.sum',
    'subscriptions.send_next_message.http_error.400.sum',
    'subscriptions.send_next_message.http_error.401.sum',
    'subscriptions.send_next_message.http_error.403.sum',
    'subscriptions.send_next_message.http_error.404.sum',
    'subscriptions.send_next_message.http_error.500.sum',
    'subscriptions.send_next_message.timeout.sum',
]
METRICS_SCHEDULED_TASKS = [
    'fire_active_last',
    'fire_created_last',
    'fire_broken_last',
    'fire_completed_last',
    'fire_incomplete_last',
    'fire_messagesets_tasks',
    'fire_week_estimate_last'
]

CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json']
CELERY_IGNORE_RESULT = True

djcelery.setup_loader()

STAGE_BASED_MESSAGING_URL = os.environ.get("STAGE_BASED_MESSAGING_URL", None)

SCHEDULER_URL = os.environ.get("SCHEDULER_URL", None)
SCHEDULER_API_TOKEN = os.environ.get("SCHEDULER_API_TOKEN", "REPLACEME")
SCHEDULER_INBOUND_API_TOKEN = \
    os.environ.get("SCHEDULER_INBOUND_API_TOKEN", "REPLACEMEONLOAD")

IDENTITY_STORE_URL = os.environ.get("IDENTITY_STORE_URL", None)
IDENTITY_STORE_TOKEN = os.environ.get("IDENTITY_STORE_TOKEN", "REPLACEME")

MESSAGE_SENDER_URL = os.environ.get("MESSAGE_SENDER_URL", None)
MESSAGE_SENDER_TOKEN = os.environ.get("MESSAGE_SENDER_TOKEN", "REPLACEME")

METRICS_URL = os.environ.get("METRICS_URL", None)
METRICS_AUTH = (
    os.environ.get("METRICS_AUTH_USER", "REPLACEME"),
    os.environ.get("METRICS_AUTH_PASSWORD", "REPLACEME"),
)

DEFAULT_REQUEST_TIMEOUT = float(os.environ.get("DEFAULT_REQUEST_TIMEOUT", 30))



from .models import Schedule, MessageSet, Message, BinaryContent
from rest_framework.viewsets import ModelViewSet
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.views import APIView
from rest_framework.response import Response
from .serializers import (ScheduleSerializer, MessageSetSerializer,
                          MessageSerializer, BinaryContentSerializer,
                          MessageListSerializer, MessageSetMessagesSerializer)


class IdCursorPagination(CursorPagination):
    ordering = "id"


class ScheduleViewSet(ModelViewSet):

    """
    API endpoint that allows Schedule models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Schedule.objects.all()
    serializer_class = ScheduleSerializer
    pagination_class = IdCursorPagination


class MessageSetViewSet(ModelViewSet):

    """
    API endpoint that allows MessageSet models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = MessageSet.objects.all()
    serializer_class = MessageSetSerializer
    filter_fields = ('short_name', 'content_type', )
    pagination_class = IdCursorPagination


class MessageViewSet(ModelViewSet):

    """
    API endpoint that allows Message models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Message.objects.all()
    serializer_class = MessageSerializer
    filter_fields = ('messageset', 'sequence_number', 'lang', )
    pagination_class = IdCursorPagination


class BinaryContentViewSet(ModelViewSet):

    """
    API endpoint that allows BinaryContent models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = BinaryContent.objects.all()
    serializer_class = BinaryContentSerializer
    pagination_class = IdCursorPagination


class MessagesContentView(ModelViewSet):

    """
    A simple ViewSet for viewing more detailed message content.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Message.objects.all()
    serializer_class = MessageListSerializer


class MessagesetMessagesContentView(ModelViewSet):

    """
    API endpoint that allows MessageSet models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = MessageSet.objects.all()
    serializer_class = MessageSetMessagesSerializer


class MessagesetLanguageView(APIView):

    """
    GET - returns languages per message set
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        data = {}
        for messageset in MessageSet.objects.all():
            data[messageset.id] = messageset.messages.order_by(
                'lang').values_list('lang', flat=True).distinct()

        return Response(data, status=status)



import codecs
import os
import re

from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):  # Stolen from txacme
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


version = get_version('seed_stage_based_messaging')


setup(
    name="seed-stage-based-messaging",
    version=version,
    url='http://github.com/praekelt/seed-stage-based-messaging',
    license='BSD',
    description='Seed Stage Based Messaging microservice',
    long_description=read('README.rst'),
    author='Praekelt.org',
    author_email='dev@praekelt.org',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Django==1.9.12',
        'djangorestframework==3.3.2',
        'dj-database-url==0.3.0',
        'psycopg2==2.7.1',
        'raven==5.32.0',
        'django-filter==0.12.0',
        'dj-static==0.0.6',
        'celery==3.1.24',
        'django-celery==3.1.17',
        'redis==2.10.5',
        'pytz==2015.7',
        'requests==2.9.1',
        'drfdocs==0.0.11',
        'seed-services-client>=0.31.0',
        'croniter==0.3.13',
        'django-cache-url==1.3.1',
        'django-redis==4.7.0',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-01 12:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='messageset',
            name='content_type',
            field=models.CharField(choices=[('text', 'Text'), ('audio', 'Audio')], default='text', max_length=20),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-04-19 06:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contentstore', '0005_auto_20160912_0923'),
    ]

    operations = [
        migrations.AddField(
            model_name='messageset',
            name='channel',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='Channel'),
        ),
    ]
